# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *
from future.utils import python_2_unicode_compatible

from .maps import Map, to_shape
from .projection import project, wgs84, MOLLWEIDE
from pyproj import Proj
import math
import multiprocessing
import pyprind
import time
import traceback


class BetterBar(pyprind.ProgBar):
    def finish(self):
        if self.cnt == self.max_iter:
            return
        else:
            self.cnt = self.max_iter
            self._finish()

    def update(self, index=None):
        if index is None:
            super(pyprind.ProgBar, self).update()
        else:
            self.cnt = index
            self._print()
            self._finish()


def chunker(iterable, chunk_size):
    for i in range(0, len(iterable), chunk_size):
        yield list(iterable[i:i + chunk_size])


def write_log_message(message):
    with open("matchstick.log", "a") as f:
        f.write(message + "\n")


def matchstick(from_map, from_objs, to_map, worker_id, verbose=True):
    """Multiprocessing worker for map matching"""
    if not from_objs:
        return

    if verbose:
        write_log_message("""Starting matchstick:
    from map: %s
    from objs: %s (%s to %s)
    to map: %s
    worker id: %s""" % (from_map, len(from_objs), min(from_objs), max(from_objs), to_map, worker_id))

    results = {}
    to_map = Map(to_map)

    if verbose:
        write_log_message("%s: Loaded to map. Vector? %s" % (worker_id, to_map.vector))

    rtree_index = to_map.create_rtree_index()
    from_map = Map(from_map)

    if verbose:
        write_log_message("%s: Loaded from map. Vector? %s" % (worker_id, from_map.vector))

    skip_projection = (from_map.crs == to_map.crs) or \
        (Proj(wgs84(from_map.crs)).is_latlong() and \
         Proj(wgs84(to_map.crs)).is_latlong())

    if from_objs:
        from_gen = zip(from_objs, from_map.select(from_objs))
    else:
        from_gen = enumerate(from_map)

    for task_index, (from_index, from_obj) in enumerate(from_gen):
        try:
            geom = to_shape(from_obj['geometry'])

            if not skip_projection:
                geom = project(geom, from_map.crs, to_map.crs)

            possibles = rtree_index.intersection(geom.bounds)
            for candidate_index in possibles:
                candidate = to_map[candidate_index]
                candidate_geom = to_shape(candidate['geometry'])
                if not geom.intersects(candidate_geom):
                    continue

                intersection = geom.intersection(candidate_geom)
                if not intersection.area:
                    continue

                results[(from_index, candidate_index)] = \
                    project(intersection, to_map.crs, MOLLWEIDE).area

        except:
            if verbose:
                write_log_message(traceback.format_exc())
            raise

    return results


def areal_calculation(from_map, from_objs, worker_id, verbose=True):
    """Multiprocessing worker for areas of each object in a map"""
    if not from_objs:
        return

    if verbose:
        write_log_message("""Starting areal calculation:
    from map: %s
    from objs: %s (%s to %s)
    worker id: %s""" % (from_map, len(from_objs), min(from_objs), max(from_objs), worker_id))

    results = {}

    from_map = Map(from_map)

    if verbose:
        write_log_message("%s: Loaded from map. Vector? %s" % (worker_id, from_map.vector))

    if from_objs:
        from_gen = zip(from_objs, from_map.select(from_objs))
    else:
        from_gen = enumerate(from_map)

    for task_index, (from_index, from_obj) in enumerate(from_gen):
        try:
            geom = to_shape(from_obj['geometry'])

            results[from_index] = \
                project(geom, from_map.crs, MOLLWEIDE).area

        except:
            if verbose:
                write_log_message(traceback.format_exc())
            raise

    return results


def matchmaker(from_map, to_map, from_objs=None, cpus=None):
    if from_objs:
        map_size = len(from_objs)
        ids = from_objs
    else:
        map_size = len(Map(from_map))
        ids = range(map_size)

    # Want a reasonable chunk size
    # But also want a maximum of 200 jobs
    # Both numbers picked more or less at random...
    chunk_size = int(max(20, map_size / 200))
    num_jobs = int(math.ceil(map_size / float(chunk_size)))

    bar = BetterBar(map_size)
    results = {}

    def areas_callback_func(data):
        results.update(data)
        bar.update(len(results))

    def intersections_callback_func(data):
        results.update(data)
        bar.update(len({key[0] for key in results}))

    if to_map is None:
        with multiprocessing.Pool(cpus or multiprocessing.cpu_count()) as pool:
            arguments = [
                (from_map, chunk, index)
                for index, chunk in enumerate(chunker(ids, chunk_size))
            ]

            function_results = []

            for argument_set in arguments:
                function_results.append(pool.apply_async(
                    areal_calculation,
                    argument_set,
                    callback=areas_callback_func
                ))
            for fr in function_results:
                fr.wait()

            if any(not fr.successful() for fr in function_results):
                raise ValueError("Couldn't complete Pandarus task")

    else:
        with multiprocessing.Pool(cpus or multiprocessing.cpu_count()) as pool:
            arguments = [
                (from_map, chunk, to_map, index)
                for index, chunk in enumerate(chunker(ids, chunk_size))
            ]

            function_results = []

            for argument_set in arguments:
                function_results.append(pool.apply_async(
                    matchstick,
                    argument_set,
                    callback=intersections_callback_func
                ))
            for fr in function_results:
                fr.wait()

            if any(not fr.successful() for fr in function_results):
                raise ValueError("Couldn't complete Pandarus task")

    bar.finish()

    return results



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

import fiona
from fiona import crs as fiona_crs
import os


wgs84 = fiona_crs.from_string("+datum=WGS84 +ellps=WGS84 +no_defs +proj=longlat")
schema = {
    'geometry': 'Polygon',
    'properties': {
        u'name': 'str:22',
    }
}


def create_record(name, coords):
    return {
        'geometry': {'coordinates': coords, 'type': 'Polygon'},
        'properties': {'name': name}
    }


def create_test_file(filepath, records, driver='GeoJSON', crs=wgs84):
    assert not os.path.exists(filepath)
    with fiona.drivers():
        with fiona.open(filepath, 'w', driver=driver, crs=crs, schema=schema) as outfile:
            for record in records:
                outfile.write(record)


def create_box(x, y, width, height):
    return [[
        (x, y),
        (x, y + height),
        (x + width, y + height),
        (x + width, y),
        (x, y)
    ]]


def create_grid(start_x=0.0, start_y=0.0, cols=10, rows=10, width=1.0,
        height=1.0):
    data = []
    for x_increment in range(cols):
        for y_increment in range(rows):
            data.append(create_box(
                start_x + x_increment * width,
                start_y + y_increment * height,
                width,
                height,
                ))
    return data


def create_grid_file(x, y, cols, rows, filepath):
    names = ["grid-cell(%s, %s)" % (i, j)
        for i in range(cols)
        for j in range(rows)
    ]
    cells = create_grid(x, y, cols, rows)
    create_test_file(
        filepath,
        [create_record(name, coords) for name, coords in zip(names, cells)]
    )



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *
from future.utils import python_2_unicode_compatible

from osgeo import ogr, gdal, osr
from osgeo.gdalconst import GDT_Float32
from shapely.wkt import loads
import os


@python_2_unicode_compatible
class Raster(object):
    """Nicer wrapper around GDAL functionality"""
    def __init__(self, filepath, band=1, loc_cache=True):
        """Load raster file"""
        assert os.path.exists(filepath), "Can't load file {}".format(filepath)
        self.filepath = filepath
        self.raster = gdal.Open(filepath)
        self.band_n = band
        self.info()
        if loc_cache:
            self.loc_cache = self.create_loc_cache()

    def _(self, band):
        if isinstance(band, int):
            return band
        else:
            return self.band_n

    def create_loc_cache(self):
        no_data = self.no_data_value()
        array = self.array()
        data = [
            (row, col)
            for col in range(self.nx)
            for row in range(self.ny)
            if array[row, col] != no_data
        ]
        return {index: obj for index, obj in enumerate(data)}

    def __repr__(self):
        fmt = "Raster(extent=({:.4e}, {:.4e}, {:.4e}, {:.4e})"
        fmt += ", pixel=({}, {}), gridsize=({}, {}))"
        return fmt.format(self.left, self.bottom, self.right, self.top,
                      self.xsize, self.ysize, self.nx, self.ny)

    def _label(self, x, y):
        return 'Cell({:.8e}, {:.8e})'.format(x, y)

    def __getitem__(self, index):
        if hasattr(self, "loc_cache"):
            if not isinstance(index, int):
                raise ValueError("Item index must be integer")
            row, col = self.loc_cache[index]
            wkt = self.wkt_for_coords(col, row)
            shape = loads(wkt)
            return {
                'geometry': wkt,
                'label': self._label(shape.centroid.x, shape.centroid.y),
                'row': row,
                'col': col,
                'value': float(self.array()[row, col]),
            }
        else:
            raise NotImplemented

    def __iter__(self):
        """Create generator that iterates over raster cells with real values.

        For each cell, return the following data structure:

        .. code-block:: python

            {
                'geometry': WKT for this raster cell geometry,
                'label': String of form "Cell(x, y)", where x and y are longitude and latitude of the raster cell centroid,
                'row': Integer row index into raster array,
                'col': Integer column index into raster array
            }
        """
        no_data = self.no_data_value()
        array = self.array()
        for col in range(self.nx):
            for row in range(self.ny):
                if array[row, col] == no_data:
                    continue
                wkt = self.wkt_for_coords(col, row)
                shape = loads(wkt)
                yield {
                    'geometry': wkt,
                    'label': self._label(shape.centroid.x, shape.centroid.y),
                    'row': row,
                    'col': col,
                    'value': float(array[row, col]),
                }

    def __len__(self):
        """Return number of raster cells with real values (no NODATA)"""
        return int((self.array() != self.no_data_value()).sum())

    @property
    def crs(self):
        """Use OSR to get coordinate reference system as PROJ.4 string"""
        geogcs = self.raster.GetProjection()
        sr = osr.SpatialReference()
        sr.ImportFromWkt(geogcs)
        return sr.ExportToProj4()

    def info(self):
        """Set raster geotransform data. Doesn't return anything"""
        t = self.raster.GetGeoTransform()
        self.left  = t[0]
        self.xsize = t[1]
        self.top   = t[3]
        self.ysize = t[5]

        self.nx = self.raster.RasterXSize
        self.ny = self.raster.RasterYSize

        self.bottom = self.top  + self.ysize * self.ny
        self.right  = self.left + self.xsize * self.nx

        assert self.right > self.left, "bounds are messed up"
        if self.bottom > self.top:
            self.bottom, self.top = self.top, self.bottom
            self.ysize *= -1

    @property
    def extent(self):
        return (self.left, self.bottom, self.right, self.top)

    def array(self, band=None):
        """Return raster data as NumPy array"""
        return self.band(band).ReadAsArray()

    def band(self, band=None):
        """Get GDAL `Band` object"""
        return self.raster.GetRasterBand(self._(band))

    def no_data_value(self, band=None):
        return self.band(band).GetNoDataValue()

    def get_writer(self, filepath):
        """Get TIFF GDAL writer with projection and geotransform of current raster.

        Assumes 1-band raster with 32 bit floating point number."""
        driver = gdal.GetDriverByName('GTiff')
        out = driver.Create(filepath, self.nx, self.ny, 1, GDT_Float32, [])
        out.SetProjection(self.raster.GetProjection())
        out.SetGeoTransform(self.raster.GetGeoTransform())
        return out

    def write_modified_array(self, filepath, array, options=[],
            nodata=None, band=None):
        """Write NumPy array `array` to `filepath`"""
        out = self.get_writer(filepath)
        new_band = out.GetRasterBand(self._(band))
        if nodata:
            new_band.SetNoDataValue(float(nodata))
        new_band.WriteArray(array)
        new_band.FlushCache()
        new_band = None
        out = None

    def wkt_for_coords(self, x, y, rounded=False):
        """Get WKT for column `x` and row `y` (integer indices).

        `rounded` rounds to 5 decimal places."""
        POLYGON_TEMPLATE = "POLYGON(({:g} {:g},{:g} {:g},{:g} {:g},{:g} {:g},{:g} {:g}))"
        if rounded:
            POLYGON_TEMPLATE = ("POLYGON(({:.5f} {:.5f},{:.5f} {:.5f},{:.5f} "
                                "{:.5f},{:.5f} {:.5f},{:.5f} {:.5f}))")
        return POLYGON_TEMPLATE.format(
            self.left + x * self.xsize, # Lower left
            self.top + y * self.ysize,
            self.left + x * self.xsize, # Upper left
            self.top + (y + 1) * self.ysize,
            self.left + (x + 1) * self.xsize, # Upper right
            self.top + (y + 1) * self.ysize,
            self.left + (x + 1) * self.xsize, # Lower right
            self.top + y * self.ysize,
            self.left + x * self.xsize, # Lower left
            self.top + y * self.ysize,
            )



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from functools import partial
from itertools import starmap
import pyproj
import shapely.geometry as sg

# License text as some code is copied directly
# Copyright (c) 2007, Sean C. Gillies
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Sean C. Gillies nor the names of
#       its contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

WGS84 = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
# See also http://spatialreference.org/ref/esri/54009/
# and http://cegis.usgs.gov/projection/pdf/nmdrs.usery.prn.pdf
MOLLWEIDE = "+proj=moll +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs"


def wgs84(s):
    """Fix no CRS or fiona giving abbreviated wgs84 definition."""
    if not s:
        return WGS84
    elif s == "+no_defs":
        return WGS84
    else:
        return s


def project(geom, from_proj=None, to_proj=None):
    """
Project a ``shapely`` geometry, and returns a new geometry of the same type from the transformed coordinates.

Default input projection is `WGS84 <https://en.wikipedia.org/wiki/World_Geodetic_System>`_, default output projection is `Mollweide <http://spatialreference.org/ref/esri/54009/>`_.

Borrowed and adapted from `mfussenegger <https://github.com/mfussenegger/Shapely/>`_.

Inputs:
    *geom*: A ``shapely`` geometry.
    *from_proj*: A ``PROJ4`` string. Optional.
    *to_proj*: A ``PROJ4`` string. Optional.

Returns:
    A ``shapely`` geometry.

    """
    from_proj = wgs84(from_proj)
    to_proj = wgs84(to_proj)
    projection_func = partial(
        pyproj.transform,
        pyproj.Proj(from_proj),
        pyproj.Proj(to_proj)
    )

    if geom.type in ('Point', 'LineString', 'Polygon'):
        # Applies the tranformation function to the coordinate list of the
        # geometry
        transform_coords = lambda geom: projection_func(*zip(*geom.coords))

        # First we try to apply func to x, y, z sequences. When func is
        # optimized for sequences, this is the fastest, though zipping
        # the results up to go back into the geometry constructors adds
        # extra cost.
        try:
            if geom.type in ('Point', 'LineString'):
                return type(geom)(list(zip(*transform_coords(geom))))
            elif geom.type == 'Polygon':
                shell = type(geom.exterior)(
                    list(zip(*transform_coords(geom.exterior))))
                holes = list(
                    type(ring)(list(zip(*transform_coords(ring))))
                               for ring in geom.interiors)
                return type(geom)(shell, holes)

        # A func that assumes x, y, z are single values will likely raise a
        # TypeError, in which case we'll try again.
        except TypeError:
            ## Rather than applying f to the list, apply it individually
            ## to each coordinate
            transform_coords = lambda geom: list(starmap(projection_func, geom.coords))

            if geom.type in ('Point', 'LineString'):
                return type(geom)(transform_coords(geom))
            elif geom.type == 'Polygon':
                shell = type(geom.exterior)(transform_coords(geom.exterior))
                if not geom.interiors:
                    holes = []
                else:
                    ring_type = type(next(geom.interiors))
                    holes = list(type(ring_type)(
                            [transform_coords(ring) for ring in geom.interiors]))
                return type(geom)(shell, holes)

    elif geom.type.startswith('Multi'):
        coords = [
            project(part, from_proj, to_proj)
            for part in geom.geoms
            # if isinstance(part, sg.polygon.Polygon)
        ]
        return type(geom)(coords)
    elif isinstance(geom, sg.collection.GeometryCollection):
        # GeometryCollection can't be instantiated with data
        # must add data iteratively
        gc = sg.collection.GeometryCollection()
        for part in geom.geoms:
            gc = gc.union(project(part, from_proj, to_proj))
        return gc
    else:
        raise ValueError('Type {} not recognized'.format(geom.type))



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *
from future.utils import python_2_unicode_compatible

from .raster import Raster
from fiona import crs as fiona_crs
from shapely.geometry import shape
from shapely.wkt import loads
import fiona
import hashlib
import rtree


class DuplicateFieldID(Exception):
    """Field ID value is duplicated and should be unique"""
    pass


def to_shape(obj):
    if isinstance(obj, str):
        # GDAL (raster)
        return loads(obj)
    elif isinstance(obj, dict):
        # Fiona (vector)
        return shape(obj)
    else:
        raise ValueError


def sha256(filepath, blocksize=65536):
    """Generate SHA 256 hash for file at `filepath`"""
    hasher = hashlib.sha256()
    fo = open(filepath, 'rb')
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()


class Map(object):
    """Common interface to raster and vector data."""
    def __init__(self, filepath, **kwargs):
        """Define the map data.

        Requires an absolute filepath. Zipped shapefiles must be specified like ``/absolute/path/to/file.zip``, and require a layer name that starts with ``/``, e.g. ``/layername``.

        Additional metadata can be provided in `kwargs`:
            * `layer` specifies the shapefile layer
            * `band` specifies the raster band

        .. warning:: The Fiona field ``id`` is not used, as there are no real constraints on these values or values types (see `Fiona manual <http://toblerity.org/fiona/manual.html#record-id>`_), and real world data is often dirty and inconsistent. Instead, we use ``enumerate`` and integer indices.

        """
        self.filepath = filepath
        self.metadata = kwargs
        self.file = self.load()
        if self.file is None:
            raise ValueError(u"Filepath '%s' can't be loaded as vector or raster" % filepath)

    def load(self):
        """Try to load vector file, fall back to raster file"""
        try:
            self.vector, self.raster = True, False
            return self.load_vector()
        except:
            self.vector, self.raster = False, True
            return Raster(self.filepath, band=self.metadata.get("band") or 1) # `band` can't be 0

    def load_vector(self):
        """Load vector file using ``fiona``."""
        if self.filepath[:-4].lower() == ".zip":
            assert self.metadata.get("layer", None), \
                "Need layer name for zipped geodata file"
            assert self.metadata['layer'][0] == "/", \
                "First character of layer name in zipped geodata must be '/'"
            with fiona.drivers():
                shapefile = fiona.open(
                    self.metadata['layer'],
                    encoding=self.metadata.get('encoding', None),
                    vfs=u"zip://" + self.filepath
                )
            return shapefile
        else:
            with fiona.drivers():
                shapefile = fiona.open(
                    self.filepath,
                    encoding=self.metadata.get('encoding', None)
                )
            return shapefile

    def select(self, obj_indices):
        """Return a generator for a selection of geographic objects.

        Integer indices for vector files follow the vector file ordering, and start at zero."""
        obj_indices = set(obj_indices)
        for index, obj in enumerate(self):
            if index in obj_indices:
                yield obj
        raise StopIteration

    def create_rtree_index(self):
        """Create `rtree <http://toblerity.org/rtree/>`_ index for efficient spatial querying."""
        self.rtree_index = rtree.Rtree()
        for index, record in enumerate(self):
            self.rtree_index.add(
                index,
                to_shape(record['geometry']).bounds
            )
        return self.rtree_index

    def get_fieldnames_dictionary(self, fieldname):
        """Get dictionary of map indices and fieldnames, like {1: 'Ghana'}"""
        if self.vector:
            return self.get_fieldnames_vector(fieldname)
        else:
            return {index: obj['label'] for index, obj in enumerate(self.file)}

    def get_fieldnames_vector(self, fieldname):
        fieldname = fieldname or self.metadata.get('field', None)
        assert fieldname, "No field name given or in metadata"
        assert fieldname in self.file.next()['properties'], \
            "Given fieldname not in file"
        fd = {index: obj['properties'].get(fieldname, None) \
            for index, obj in enumerate(self)}
        if len(fd.keys()) != len(set(fd.values())):
            raise DuplicateFieldID(
                "Given field name not unique for all records"
            )
        return fd

    @property
    def hash(self):
        return sha256(self.filepath)

    @property
    def info(self):
        print("info!!!")

    @property
    def crs(self):
        """Coordinate reference system, as defined by vector file."""
        if self.vector:
            return fiona_crs.to_string(self.file.crs)
        else:
            return self.file.crs

    def __iter__(self):
        return iter(self.file)

    def __getitem__(self, index):
        """Incredibly awkward, but avoids loading dataset into memory."""
        if self.raster and hasattr(self.file, "loc_cache"):
            return self.file[index]
        for i, obj in enumerate(self):
            if i == index:
                return obj

    def __len__(self):
        return len(self.file)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

import sys
if sys.version_info < (3, 0):
    from ._unicodecsv import writer
else:
    from csv import writer


def csv_exporter(data, metadata, filepath, **kwargs):
    w = writer(open(filepath, "w", encoding='utf-8'))
    for line in data:
        w.writerow(line)
    return filepath



#!/usr/bin/env python
# encoding: utf-8
"""Pandarus command line controller.

This program does one of:
* Calculates the area of each spatial unit in a map.
* Match two geospatial datasets, and calculates the area of each intersecting spatial unit in map1 and map2.

Usage:
  pandarus area <map> [--field=<field>|--band=<band>] <output> [csv|json|pickle] [options]
  pandarus intersect <map1> [--field1=<field1>|--band1=<band1>] <map2> [--field2=<field2>|--band2=<band2>] <output> [csv|json|pickle] [options]

Options:
  --no-bz2          Don't compress output with BZip2
  --cpus=<cpus>     Number of cpus to use (default is all)
  --no-projection   Compute areas in CRS of `map2`, don't use Mollweide projection
  -h --help         Show this screen.
  --version         Show version.

"""
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from docopt import docopt
from pandarus import *
import os
import sys


def main():
    arguments = docopt(__doc__, version='Pandarus CLI 0.3')
    # if arguments['info']:
    #     print(Map(arguments['<map>']).info)
    if arguments['area']:
        format = ('csv' if arguments['csv'] else 0) or ('pickle' if arguments['pickle'] else 0) or 'json'
        output = os.path.abspath(arguments['<output>']) + '.' + format

        # Don't overwrite existing output
        if os.path.exists(output) or os.path.exists(output + ".bz2"):
            sys.exit("ERROR: Output file `%s` already exists" % output)

        kwargs = {
            'from_filepath': arguments['<map>'],
        }
        if arguments.get('--field'):
            kwargs['from_metadata'] = {'field': arguments['--field']}
        elif arguments.get('--band'):
            kwargs['from_metadata'] = {'band': arguments['--band']}

        controller = Pandarus(**kwargs)
        controller.areas(cpus=int(arguments['--cpus'] or 0))

        if kwargs.get('from_metadata') or controller.from_map.raster:
            controller.add_areas_map_fieldname()

        exported = controller.export(
            output,
            format,
            compress=not arguments['--no-bz2']
        )
        print("Finished Pandarus areas job; created `%s`" % exported)
    elif arguments['intersect']:
        format = ('csv' if arguments['csv'] else 0) or ('pickle' if arguments['pickle'] else 0) or 'json'
        output = os.path.abspath(arguments['<output>']) + '.' + format

        # Don't overwrite existing output
        if os.path.exists(output) or os.path.exists(output + ".bz2"):
            sys.exit("ERROR: Output file `%s` already exists" % output)

        kwargs = {
            'from_filepath': arguments['<map1>'],
            'to_filepath': arguments['<map2>'],
        }
        if arguments.get('--field1'):
            kwargs['from_metadata'] = {'field': arguments['--field1']}
        elif arguments.get('--band1'):
            kwargs['from_metadata'] = {'band': arguments['--band1']}
        if arguments.get('--field2'):
            kwargs['to_metadata'] = {'field': arguments['--field2']}
        elif arguments.get('--band2'):
            kwargs['from_metadata'] = {'band': arguments['--band2']}

        controller = Pandarus(**kwargs)
        controller.match(cpus=int(arguments['--cpus'] or 0))

        if kwargs.get('from_metadata') or controller.from_map.raster:
            controller.add_from_map_fieldname()
        if kwargs.get('to_metadata') or controller.to_map.raster:
            controller.add_to_map_fieldname()

        exported = controller.export(
            output,
            format,
            compress=not arguments['--no-bz2']
        )
        print("Finished Pandarus intersection job; created `%s`" % exported)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .maps import Map
from .matching import matchmaker
from .export import json_exporter, csv_exporter, pickle_exporter
import os


export_functions = {
    "json": json_exporter,
    "csv": csv_exporter,
    "pickle": pickle_exporter
}


class Pandarus(object):
    """Controller for all actions."""
    def __init__(self, from_filepath, to_filepath=None,
            from_metadata={}, to_metadata={}):
        self.from_map = Map(from_filepath, **from_metadata)
        self.metadata = {'first': {
            'sha256': self.from_map.hash,
            'filename': os.path.basename(from_filepath)
        }}
        self.metadata['first'].update(from_metadata)
        if to_filepath is not None:
            self.to_map = Map(to_filepath, **to_metadata)
            self.metadata['second'] = {
                'sha256': self.to_map.hash,
                'filename': os.path.basename(to_filepath)
            }
            self.metadata['second'].update(to_metadata)

    def match(self, cpus=None):
        self.data = matchmaker(
            self.from_map.filepath,
            self.to_map.filepath,
            cpus=cpus,
        )
        return self.data

    def areas(self, cpus=None):
        self.data = matchmaker(
            self.from_map.filepath,
            None,
            cpus=cpus,
        )
        return self.data

    def add_from_map_fieldname(self, fieldname=None):
        """Turn feature integer indices into actual field values using field `fieldname`"""
        if not self.data:
            raise ValueError("Must match maps first")
        mapping_dict = self.from_map.get_fieldnames_dictionary(fieldname)
        self.data = {
            (mapping_dict[k[0]], k[1]): v
            for k, v in self.data.items()
        }

    def add_to_map_fieldname(self, fieldname=None):
        """Turn feature integer indices into actual field values using field `fieldname`"""
        if not self.data:
            raise ValueError("Must match maps first")
        mapping_dict = self.to_map.get_fieldnames_dictionary(fieldname)
        self.data = {
            (k[0], mapping_dict[k[1]]): v
            for k, v in self.data.items()
        }

    def add_areas_map_fieldname(self, fieldname=None):
        """Turn feature integer indices into actual field values using field `fieldname`"""
        if not self.data:
            raise ValueError("Must match maps first")
        mapping_dict = self.from_map.get_fieldnames_dictionary(fieldname)
        self.data = {mapping_dict[k]: v for k, v in self.data.items()}

    def export(self, filepath, kind="json", compress=True):
        if filepath[-len(kind):].lower() != kind:
            filepath = filepath + '.' + kind
        exporter = export_functions[kind]
        if kind in {'csv', 'json'} and isinstance(list(self.data.keys())[0], tuple):
            self.unpack_tuples()
        else:
            self.unpack_dictionary()
        return exporter(self.data, self.metadata, filepath, compressed=compress)

    def unpack_tuples(self):
        self.data = [(k[0], k[1], v) for k, v in self.data.items()]

    def unpack_dictionary(self):
        self.data = [(k, v) for k, v in self.data.items()]



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

import bz2
import codecs
import json


def json_exporter(data, metadata, filepath, compressed=True):
    if compressed:
        filepath += ".bz2"
        with bz2.BZ2File(filepath, "w") as f:
            f.write(json.dumps({'data': data, 'metadata': metadata},
                ensure_ascii=False).encode('utf-8'))
    else:
        with codecs.open(filepath, "w", encoding="utf-8") as f:
            json.dump({'data': data, 'metadata': metadata}, f, ensure_ascii=False)
    return filepath



from setuptools import setup

setup(
    version="0.6",
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    package_data={'pandarus': [
        "tests/data/*.gpkg",
        "tests/data/*.tif",
    ]},
    entry_points = {
        'console_scripts': [
            'pandarus = pandarus.cli:main',
        ]
    },
    install_requires=[
        "docopt",
        "eight",
        "fiona",
        "GDAL",
        "psutil",
        "pyprind",
        "pyproj",
        "Rtree",
        "shapely",
    ],
    license=open('LICENSE.txt').read(),
    long_description=open('README.rst').read(),
    name='pandarus',
    packages=["pandarus", "pandarus.tests", "pandarus.export"],
    url="https://bitbucket.org/cmutel/pandarus",
)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .json_exporter import json_exporter
from .csv_exporter import csv_exporter

try:
    import cPicle as pickle
except ImportError:
    import pickle


def pickle_exporter(data, metadata, filepath, **kwargs):
    with open(filepath, "wb") as f:
        pickle.dump({'data': data, 'metadata': metadata}, f)
    return filepath



# # -*- coding: utf-8 -*-
# Copyright 2010 Jeremy Dunck. All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:

#    1. Redistributions of source code must retain the above copyright notice, this list of
#       conditions and the following disclaimer.

#    2. Redistributions in binary form must reproduce the above copyright notice, this list
#       of conditions and the following disclaimer in the documentation and/or other materials
#       provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY JEREMY DUNCK ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JEREMY DUNCK OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those of the
# authors and should not be interpreted as representing official policies, either expressed
# or implied, of Jeremy Dunck.

# unicodecsv 0.9.4
# https://github.com/jdunck/python-unicodecsv

import csv
from itertools import izip

#http://semver.org/
VERSION = (0, 9, 4)
__version__ = ".".join(map(str,VERSION))

pass_throughs = [
    'register_dialect',
    'unregister_dialect',
    'get_dialect',
    'list_dialects',
    'field_size_limit',
    'Dialect',
    'excel',
    'excel_tab',
    'Sniffer',
    'QUOTE_ALL',
    'QUOTE_MINIMAL',
    'QUOTE_NONNUMERIC',
    'QUOTE_NONE',
    'Error'
]
__all__ = [
    'reader',
    'writer',
    'DictReader',
    'DictWriter',
] + pass_throughs

for prop in pass_throughs:
    globals()[prop]=getattr(csv, prop)

def _stringify(s, encoding, errors):
    if s is None:
        return ''
    if isinstance(s, unicode):
        return s.encode(encoding, errors)
    elif isinstance(s, (int , float)):
        pass #let csv.QUOTE_NONNUMERIC do its thing.
    elif not isinstance(s, str):
        s=str(s)
    return s

def _stringify_list(l, encoding, errors='strict'):
    try:
        return [_stringify(s, encoding, errors) for s in iter(l)]
    except TypeError, e:
        raise csv.Error(str(e))

def _unicodify(s, encoding):
    if s is None:
        return None
    if isinstance(s, (unicode, int, float)):
        return s
    elif isinstance(s, str):
        return s.decode(encoding)
    return s


class UnicodeWriter(object):
    """
    >>> import unicodecsv
    >>> from cStringIO import StringIO
    >>> f = StringIO()
    >>> w = unicodecsv.writer(f, encoding='utf-8')
    >>> w.writerow((u'é', u'ñ'))
    >>> f.seek(0)
    >>> r = unicodecsv.reader(f, encoding='utf-8')
    >>> row = r.next()
    >>> row[0] == u'é'
    True
    >>> row[1] == u'ñ'
    True
    """
    def __init__(self, f, dialect=csv.excel, encoding='utf-8', errors='strict',
                 *args, **kwds):
        self.encoding = encoding
        self.writer = csv.writer(f, dialect, *args, **kwds)
        self.encoding_errors = errors

    def writerow(self, row):
        self.writer.writerow(_stringify_list(row, self.encoding, self.encoding_errors))

    def writerows(self, rows):
        for row in rows:
          self.writerow(row)

    @property
    def dialect(self):
        return self.writer.dialect
writer = UnicodeWriter


class UnicodeReader(object):
    def __init__(self, f, dialect=None, encoding='utf-8', errors='strict',
                 **kwds):
        format_params = ['delimiter', 'doublequote', 'escapechar', 'lineterminator', 'quotechar', 'quoting', 'skipinitialspace']
        if dialect is None:
            if not any([kwd_name in format_params for kwd_name in kwds.keys()]):
                dialect = csv.excel
        self.reader = csv.reader(f, dialect, **kwds)
        self.encoding = encoding
        self.encoding_errors = errors

    def next(self):
        row = self.reader.next()
        encoding = self.encoding
        encoding_errors = self.encoding_errors
        float_ = float
        unicode_ = unicode
        return [(value if isinstance(value, float_) else
                 unicode_(value, encoding, encoding_errors)) for value in row]

    def __iter__(self):
        return self

    @property
    def dialect(self):
        return self.reader.dialect

    @property
    def line_num(self):
        return self.reader.line_num
reader = UnicodeReader


class DictWriter(csv.DictWriter):
    """
    >>> from cStringIO import StringIO
    >>> f = StringIO()
    >>> w = DictWriter(f, ['a', u'ñ', 'b'], restval=u'î')
    >>> w.writerow({'a':'1', u'ñ':'2'})
    >>> w.writerow({'a':'1', u'ñ':'2', 'b':u'ø'})
    >>> w.writerow({'a':u'é', u'ñ':'2'})
    >>> f.seek(0)
    >>> r = DictReader(f, fieldnames=['a', u'ñ'], restkey='r')
    >>> r.next() == {'a': u'1', u'ñ':'2', 'r': [u'î']}
    True
    >>> r.next() == {'a': u'1', u'ñ':'2', 'r': [u'\xc3\xb8']}
    True
    >>> r.next() == {'a': u'\xc3\xa9', u'ñ':'2', 'r': [u'\xc3\xae']}
    True
    """
    def __init__(self, csvfile, fieldnames, restval='', extrasaction='raise', dialect='excel', encoding='utf-8', errors='strict', *args, **kwds):
        self.encoding = encoding
        csv.DictWriter.__init__(self, csvfile, fieldnames, restval, extrasaction, dialect, *args, **kwds)
        self.writer = UnicodeWriter(csvfile, dialect, encoding=encoding, errors=errors, *args, **kwds)
        self.encoding_errors = errors

    def writeheader(self):
        fieldnames = _stringify_list(self.fieldnames, self.encoding, self.encoding_errors)
        header = dict(zip(self.fieldnames, self.fieldnames))
        self.writerow(header)


class DictReader(csv.DictReader):
    """
    >>> from cStringIO import StringIO
    >>> f = StringIO()
    >>> w = DictWriter(f, fieldnames=['name', 'place'])
    >>> w.writerow({'name': 'Cary Grant', 'place': 'hollywood'})
    >>> w.writerow({'name': 'Nathan Brillstone', 'place': u'øLand'})
    >>> w.writerow({'name': u'Willam ø. Unicoder', 'place': u'éSpandland'})
    >>> f.seek(0)
    >>> r = DictReader(f, fieldnames=['name', 'place'])
    >>> print(r.next() == {'name': 'Cary Grant', 'place': 'hollywood'})
    True
    >>> print(r.next() == {'name': 'Nathan Brillstone', 'place': u'øLand'})
    True
    >>> print(r.next() == {'name': u'Willam ø. Unicoder', 'place': u'éSpandland'})
    True
    """
    def __init__(self, csvfile, fieldnames=None, restkey=None, restval=None,
                 dialect='excel', encoding='utf-8', errors='strict', *args,
                 **kwds):
        if fieldnames is not None:
            fieldnames = _stringify_list(fieldnames, encoding)
        csv.DictReader.__init__(self, csvfile, fieldnames, restkey, restval, dialect, *args, **kwds)
        self.reader = UnicodeReader(csvfile, dialect, encoding=encoding,
                                    errors=errors, *args, **kwds)
        if fieldnames is None and not hasattr(csv.DictReader, 'fieldnames'):
            # Python 2.5 fieldnames workaround. (http://bugs.python.org/issue3436)
            reader = UnicodeReader(csvfile, dialect, encoding=encoding, *args, **kwds)
            self.fieldnames = _stringify_list(reader.next(), reader.encoding)
        self.unicode_fieldnames = [_unicodify(f, encoding) for f in
                                   self.fieldnames]
        self.unicode_restkey = _unicodify(restkey, encoding)

    def next(self):
        row = csv.DictReader.next(self)
        result = dict((uni_key, row[str_key]) for (str_key, uni_key) in
                      izip(self.fieldnames, self.unicode_fieldnames))
        rest = row.get(self.restkey)
        if rest:
            result[self.unicode_restkey] = rest
        return result
